@@ -3,7 +3,9 @@
 *** Type definitions ***
 
 RowVal = str | int | float
+
 Row = dict[str, RowVal]
+
     Row represents a row in a table, str is a column name (field)
 """
 
@@ -127,14 +129,84 @@ class Rows:
             slicing here is the same as list slicing except that negative values
             cannot be used
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
-
         if isinstance(key, str):
             return [r[key] for r in self._iter()]
         if isinstance(key, slice):
             return self._islice(key.start, key.stop, key.step)
         raise ValueError("Must pass a str(column name) or a slice")
 
+    def split(self):
+        """Returns a list of Rows grouped by previous method
+
+        Returns
+        -------
+        list[Rows]
+        """
+        _raise_exception_if_not_preceded_by_grouping_methods(self)
+
+        hist = self._history
+
+        pgen = hist[-1]['genfn']
+
+        if hist[-1]['cmd'] == 'group':
+            return [Rows(rows) for _, rows in pgen()]
+
+        return [Rows(rows) for rows in pgen()]
+
+    def index(self, field, start=0, step=1):
+        """Add an index column
+
+        Parameters
+        ----------
+        field: str
+        start: int (default 0)
+        step: int (default 1)
+
+        Returns
+        -------
+        Rows
+        """
+        def gen_single():
+            count = start
+
+            def gen(row):
+                nonlocal count
+                newrow = row.copy()
+                newrow[field] = count
+                count += step
+                yield newrow
+            return gen
+
+        def gen_group():
+            count = start
+
+            def gen(_key, rows):
+                nonlocal count
+                for row in rows:
+                    newrow = row.copy()
+                    newrow[field] = count
+                    yield newrow
+                count += step
+            return gen
+
+        def gen_windowed():
+            count = start
+
+            def gen(rows):
+                nonlocal count
+                for row in rows:
+                    newrow = row.copy()
+                    newrow[field] = count
+                    yield newrow
+                count += step
+
+            return gen
+
+        return self._apply(gen4single=gen_single(),
+                           gen4group=gen_group(),
+                           gen4windowed=gen_windowed()
+                           )
+
     def zip(self, other):
         """Attach two Rows side by side until either one is depleted
 
@@ -253,7 +325,7 @@ class Rows:
     def zip_longest(self, other):
         """Attach two Rows side by side as long as any of them remained
 
-        Empty columns (with values of '') generated
+        Empty columns (with values of empty string, '') generated
 
         Parameters
         ----------
@@ -439,42 +511,6 @@ class Rows:
 
         return newself
 
-    # End-users might not be able to pass proper functions.
-    def _apply(self, func):
-        newself = copy(self)
-        # pylint: disable=protected-access
-        pgen = newself._history[-1]['genfn']
-
-        def gen_single():
-            for row in pgen():
-                yield from func(row)
-
-        def gen_group():
-            for key, rows in pgen():
-                yield from func(key, rows)
-
-        def gen_windowed():
-            for rows in pgen():
-                yield from func(rows)
-
-        genfn = None
-        # pylint: disable=protected-access
-        if newself._history[-1]['cmd'] == 'group':
-            genfn = gen_group
-        # pylint: disable=protected-access
-        elif newself._history[-1]['cmd'] == 'windowed':
-            genfn = gen_windowed
-        else:
-            genfn = gen_single
-
-        # Do not use += operator here, it modifies the object
-        # pylint: disable=protected-access
-        newself._history = newself._history + [{
-            'cmd': '_apply',
-            'genfn': genfn
-        }]
-        return newself
-
     def group(self, fields):
         """Group consecutive rows with the same values for given fields
 
@@ -524,6 +560,44 @@ class Rows:
 
         return newself
 
+    def _apply(self, gen4single=None, gen4group=None, gen4windowed=None):
+        if (not gen4group) and (not gen4windowed):
+            _raise_exception_if_preceded_by_grouping_methods(self)
+
+        if (gen4group or gen4windowed) and (not gen4single):
+            _raise_exception_if_not_preceded_by_grouping_methods(self)
+
+        newself = copy(self)
+        # pylint: disable=protected-access
+        hist = newself._history
+        pgen = hist[-1]['genfn']
+
+        genfn = None
+        if hist[-1]['cmd'] == 'group':
+            def build_gen4group():
+                for key, rows in pgen():
+                    yield from gen4group(key, rows)
+            genfn = build_gen4group
+
+        elif hist[-1]['cmd'] == 'windowed':
+            def build_gen4windowed():
+                for rows in pgen():
+                    yield from gen4windowed(rows)
+            genfn = build_gen4windowed
+        else:
+            def build_gen4single():
+                for row in pgen():
+                    yield from gen4single(row)
+            genfn = build_gen4single
+
+        # Do not use += operator here, it modifies the object
+        # pylint: disable=protected-access
+        newself._history = newself._history + [{
+            'cmd': '_apply',
+            'genfn': genfn
+        }]
+        return newself
+
     def map(self, func):
         """Applies func to each row or Rows (when grouped)
 
@@ -538,32 +612,27 @@ class Rows:
         -------
         Rows
         """
-        if self._history[-1]['cmd'] == 'group':
-            def _fn2gen(func):
-                def gen(_key, rows):
-                    val = func(Rows(rows))
-                    if isinstance(val, dict):
-                        yield val
-                    elif isinstance(val, Rows):
-                        # pylint: disable=protected-access
-                        yield from val._iter()
-                    # if it's neither dict nor Rows simply ignored.
-                    # no Nonetype check
-                return gen
-            return self._apply(_fn2gen(func))
+        def _fn2gen_group(func):
+            def gen(_key, rows):
+                val = func(Rows(rows))
+                if isinstance(val, dict):
+                    yield val
+                elif isinstance(val, Rows):
+                    # pylint: disable=protected-access
+                    yield from val._iter()
+                # if it's neither dict nor Rows simply ignored.
+                # no Nonetype check
+            return gen
 
-        if self._history[-1]['cmd'] == 'windowed':
-            def _fn2gen1(func):
-                def gen(rows):
-                    val = func(Rows(rows))
-                    if isinstance(val, dict):
-                        yield val
-                    elif isinstance(val, Rows):
-                        # pylint: disable=protected-access
-                        yield from val._iter()
-                return gen
-
-            return self._apply(_fn2gen1(func))
+        def _fn2gen_windowed(func):
+            def gen(rows):
+                val = func(Rows(rows))
+                if isinstance(val, dict):
+                    yield val
+                elif isinstance(val, Rows):
+                    # pylint: disable=protected-access
+                    yield from val._iter()
+            return gen
 
         # single row
         def _fn2gen_single(func):
@@ -575,7 +644,12 @@ class Rows:
                     # pylint: disable=protected-access
                     yield from val._iter()
             return gen
-        return self._apply(_fn2gen_single(func))
+
+        return self._apply(
+            gen4single=_fn2gen_single(func),
+            gen4group=_fn2gen_group(func),
+            gen4windowed=_fn2gen_windowed(func)
+        )
 
     # name 'by' is inevitable
     # pylint: disable=invalid-name
@@ -614,13 +688,25 @@ class Rows:
         -------
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
 
-        def gen(row):
+        def single(row):
             if pred(row):
                 yield row
 
-        return self._apply(gen)
+        def group(_key, rows):
+            rows = list(rows)
+            if pred(Rows(rows)):
+                yield from rows
+
+        def windowed(rows):
+            if pred(Rows(rows)):
+                yield from rows
+
+        return self._apply(
+            gen4single=single,
+            gen4group=group,
+            gen4windowed=windowed
+        )
 
     def update(self, **kwargs):
         """Updates each row with new ones
@@ -637,8 +723,6 @@ class Rows:
         -------
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
-
         def updatefn1(row):
             newrow = row.copy()
             for field, val in kwargs.items():
@@ -652,8 +736,8 @@ class Rows:
             yield row
 
         if self._history[0]['cmd'] == 'fetch':
-            return self._apply(updatefn2)
-        return self._apply(updatefn1)
+            return self._apply(gen4single=updatefn2)
+        return self._apply(gen4single=updatefn1)
 
     # pretty expensive for what it actually does
     # but this version does not depend on the order of rows.
@@ -672,15 +756,13 @@ class Rows:
         -------
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
-
         kwargs_rev = {oldkey: newkey for newkey, oldkey in kwargs.items()}
 
         def renamefn(row):
             yield {kwargs_rev.get(oldkey, oldkey): val
                    for oldkey, val in row.items()}
 
-        return self._apply(renamefn)
+        return self._apply(gen4single=renamefn)
 
     def fold(self, **kwargs):
         """n rows to 1 row
@@ -696,7 +778,6 @@ class Rows:
         -------
         Rows
         """
-        _raise_exception_if_not_preceded_by_grouping_methods(self)
 
         # fold must be preceded by `by`
         grouping_fields = []
@@ -719,10 +800,7 @@ class Rows:
                 row[newfield] = val(rows) if callable(val) else val
             yield row
 
-        if self._history[-1]['cmd'] == 'windowed':
-            return self._apply(foldfn1)
-        # then self._history[-1]['cmd'] must be 'group', it's check above
-        return self._apply(foldfn2)
+        return self._apply(gen4windowed=foldfn1, gen4group=foldfn2)
 
     def select(self, fields):
         """select fields discarding others
@@ -736,13 +814,12 @@ class Rows:
         Returns:
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
         fields = _listify(fields)
 
         def selectfn(row):
             yield {field: row[field] for field in fields}
 
-        return self._apply(selectfn)
+        return self._apply(gen4single=selectfn)
 
     def deselect(self, fields):
         """Deselect fields
@@ -756,7 +833,6 @@ class Rows:
         Returns:
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
         fields = _listify(fields)
 
         fields_set = set(fields)
@@ -765,7 +841,7 @@ class Rows:
             yield {field: val for field, val in row.items()
                    if field not in fields_set}
 
-        return self._apply(deselectfn)
+        return self._apply(gen4single=deselectfn)
 
     def join(self, other, join_type="inner"):
         """Joins two Rows in SQL style
@@ -863,7 +939,7 @@ class Rows:
             yield next(rows)
 
         # pylint: disable=protected-access
-        return self.by(fields)._apply(distinctfn)
+        return self.by(fields)._apply(gen4group=distinctfn)
 
     def windowed(self, chunk_size, step=1):
         """Returns a Rows of a sliding window(rows) with chunk_size skipping
@@ -883,14 +959,33 @@ class Rows:
         -------
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
 
         newself = copy(self)
         # pylint: disable=protected-access
-        pgen = newself._history[-1]['genfn']
+        hist = newself._history
 
-        def gen():
-            yield from _windowed(pgen(), chunk_size, step)
+        pgen = hist[-1]['genfn']
+
+        genfn = None
+        if hist[-1]['cmd'] == 'group':
+            def gen_group():
+                def rows_list():
+                    for _, rows in pgen():
+                        yield list(rows)
+
+                for rows_batch in _windowed(rows_list(), chunk_size, step):
+                    yield [row for rows in rows_batch for row in rows]
+            genfn = gen_group
+
+        elif hist[-1]['cmd'] == 'windowed':
+            def gen_windowed():
+                for rows_batch in _windowed(pgen(), chunk_size, step):
+                    yield [row for rows in rows_batch for row in rows]
+            genfn = gen_windowed
+        else:
+            def gen_single():
+                yield from _windowed(pgen(), chunk_size, step)
+            genfn = gen_single
 
         # Do not use += operator here, it modifies the object
         # pylint: disable=protected-access
@@ -898,23 +993,73 @@ class Rows:
             'cmd': 'windowed',
             'n': chunk_size,
             'step': step,
-            'genfn': gen
+            'genfn': genfn
         }]
         return newself
 
     def _islice(self, *args):
         newself = copy(self)
-        pgen = self._history[-1]['genfn']
+        # pylint: disable=protected-access
+        hist = self._history
+        pgen = hist[-1]['genfn']
 
-        def gen():
-            yield from islice(pgen(), *args)
+        genfn = None
+        if hist[-1]['cmd'] == 'group':
+            def gen_group():
+                for _, rows in islice(pgen(), *args):
+                    yield from rows
+            genfn = gen_group
+        elif hist[-1]['cmd'] == 'windowed':
+            def gen_windowed():
+                for rows in islice(pgen(), *args):
+                    yield from rows
+            genfn = gen_windowed
+        else:
+            def gen_single():
+                yield from islice(pgen(), *args)
+            genfn = gen_single
 
-        # Do not use += operator here, it modifies the object
+       # Do not use += operator here, it modifies the object
         # pylint: disable=protected-access
         newself._history = newself._history + [{
             'cmd': '_islice',
             'args': args,
-            'genfn': gen
+            'genfn': genfn
+        }]
+        return newself
+
+    def _take_or_drop_while(self, pred, take_or_drop, cmd_str):
+        newself = copy(self)
+        # pylint: disable=protected-access
+        hist = newself._history
+        pgen = hist[-1]['genfn']
+
+        genfn = None
+        if hist[-1]['cmd'] == 'group':
+            def gen_group():
+                for rows in take_or_drop(
+                    pred, (Rows(rows) for _, rows in pgen())
+                ):
+                    yield from rows._iter()
+            genfn = gen_group
+
+        elif hist[-1]['cmd'] == 'windowed':
+            def gen_windowed():
+                for rows in take_or_drop(
+                    pred, (Rows(rows) for rows in pgen())
+                ):
+                    yield from rows._iter()
+            genfn = gen_windowed
+        else:
+            def gen_single():
+                yield from take_or_drop(pred, pgen())
+            genfn = gen_single
+
+        # Do not use += operator here, it modifies the object
+        # pylint: disable=protected-access
+        newself._history = newself._history + [{
+            'cmd': cmd_str,
+            'genfn': genfn
         }]
         return newself
 
@@ -931,26 +1076,10 @@ class Rows:
         Rows
 
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
-
-        newself = copy(self)
-        # pylint: disable=protected-access
-        pgen = newself._history[-1]['genfn']
-
-        def gen():
-            yield from takewhile(pred, pgen())
-
-        # Do not use += operator here, it modifies the object
-        # pylint: disable=protected-access
-        newself._history = newself._history + [{
-            'cmd': 'takewhile',
-            'genfn': gen
-        }]
-        return newself
+        return self._take_or_drop_while(pred, takewhile, 'takewhile')
 
     def dropwhile(self, pred):
-        """Returns a Rows that would drop rows as long as the pred
-        is true
+        """Returns a Rows that would drop rows as long as the pred is true
 
         Parameters
         ----------
@@ -960,22 +1089,7 @@ class Rows:
         -------
         Rows
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
-
-        newself = copy(self)
-        # pylint: disable=protected-access
-        pgen = newself._history[-1]['genfn']
-
-        def gen():
-            yield from dropwhile(pred, pgen())
-
-        # Do not use += operator here, it modifies the object
-        # pylint: disable=protected-access
-        newself._history = newself._history + [{
-            'cmd': 'dropwhile',
-            'genfn': gen
-        }]
-        return newself
+        return self._take_or_drop_while(pred, dropwhile, 'dropwhile')
 
     def size(self):
         """Returns the number(size) of rows
@@ -984,15 +1098,17 @@ class Rows:
         -------
         int
         """
-        _raise_exception_if_preceded_by_grouping_methods(self)
+        hist = self._history
+        if hist[-1]['cmd'] == 'group' or hist[-1]['cmd'] == 'windowed':
+            return sum(1 for _ in self._iter())
 
         # There are rooms for this to be more efficient because some of the
         # methods do not affect the size of the Rows but
         # don't think it's worth the trouble.
-        if len(self._history) == 1 and self._history[0]['cmd'] == 'fetch':
-            origin = self._history[0]
+        if len(hist) == 1 and hist[0]['cmd'] == 'fetch':
+            origin = hist[0]
             return _get_size(origin['conn'], origin['tname'])
-        if len(self._history) == 1 and self._history[0]['cmd'] == 'read':
+        if len(hist) == 1 and hist[0]['cmd'] == 'read':
             return len(self._src)
         return sum(1 for _ in self._iter())
 
