@@ -36,6 +36,85 @@ class TestCore(unittest.TestCase):
         if Path(self.dbfile1).is_file():
             os.remove(self.dbfile1)
 
+    def test_takewhile_group(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        self.assertEqual(conn['t1'].by('col1')
+                         .takewhile(lambda rs: rs['col1'][0] < 'c').list(),
+                         [{'col1': 'a', 'col2': 4},
+                          {'col1': 'a', 'col2': 5},
+                          {'col1': 'b', 'col2': 1}]
+                         )
+
+        self.assertEqual(conn['t1'].windowed(4, 2)
+                         .takewhile(lambda rs: rs.size() == 4).list(),
+                         [{'col1': 'a', 'col2': 4},
+                          {'col1': 'a', 'col2': 5},
+                          {'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7}])
+
+    def test_dropwhile_group(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        self.assertEqual(conn['t1'].by('col1')
+                         .dropwhile(lambda rs: rs['col1'][0] < 'c').list(),
+                         [{'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7},
+                          {'col1': 'd', 'col2': 2}])
+
+        self.assertEqual(conn['t1'].windowed(4, 2)
+                         .dropwhile(lambda rs: rs.size() == 4).list(),
+                         [{'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7},
+                          {'col1': 'd', 'col2': 2}])
+
+    def test_split(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        xs = conn['t1'].group('col1').split()
+        self.assertEqual(len(xs[0].list()), 2)
+        self.assertEqual(len(xs[1].list()), 1)
+        self.assertEqual(len(xs[2].list()), 3)
+        self.assertEqual(len(xs[3].list()), 1)
+
+        xs = conn['t1'].windowed(4, 2).split()
+        self.assertEqual(len(xs[0].list()), 4)
+        self.assertEqual(len(xs[1].list()), 4)
+        self.assertEqual(len(xs[2].list()), 3)
+
+        xs1 = xs[2].update(foo=4)
+        # test xs[1] is unaffected
+        self.assertEqual(xs[1].list(),
+                         [{'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7}])
+
+    def test_size_group(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+        self.assertEqual(conn['t1'].group('col1').size(), 4)
+        self.assertEqual(conn['t1'].windowed(4, 2).size(), 3)
+
+    def test_index1(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].index('foo')
+        self.assertEqual(conn['t1_1']['foo'], [0, 1, 2, 3, 4, 5, 6])
+
+        conn['t1_1'] = conn['t1'].group('col1').index('foo', 1, 2)
+        self.assertEqual(conn['t1_1']['foo'], [1, 1, 3, 5, 5, 5, 7])
+        conn['t1_1'] = conn['t1'].windowed(4, 2).index('foo', 1, 2)
+        self.assertEqual(conn['t1_1']['foo'], [
+                         1, 1, 1, 1, 3, 3, 3, 3, 5, 5, 5])
 
     def test_order_when_no_row(self):
         rs1 = Rows([])
@@ -122,7 +201,6 @@ class TestCore(unittest.TestCase):
         conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
         self.assertEqual(conn['t1_1'].list(), conn['t1'].list())
 
-
     def test_do_later(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -422,6 +500,26 @@ class TestCore(unittest.TestCase):
         self.assertEqual(conn['t1_1_double'].list(),
                          conn['t1_1_double2'].list())
 
+    def test_islice_group(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        self.assertEqual(conn['t1'].by('col1')[1:3].list(),
+                         [{'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7}])
+
+        self.assertEqual(conn['t1'].windowed(4, 2)[1:3].list(),
+                         [{'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7},
+                          {'col1': 'd', 'col2': 2}
+                          ])
+
     def test_islice(self):
 
         conn = Conn(self.dbfile)
@@ -475,6 +573,52 @@ class TestCore(unittest.TestCase):
                              .order('col2')
                              .dropwhile(lambda r: r['col2'] < 4).list()), 4)
 
+    def test_windowed_group(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        self.assertEqual(conn['t1'].by('col1').windowed(2, 1)
+                         .index('cnt').list(),
+                         [{'col1': 'a', 'col2': 4, 'cnt': 0},
+                          {'col1': 'a', 'col2': 5, 'cnt': 0},
+                          {'col1': 'b', 'col2': 1, 'cnt': 0},
+                          {'col1': 'b', 'col2': 1, 'cnt': 1},
+                          {'col1': 'c', 'col2': 3, 'cnt': 1},
+                          {'col1': 'c', 'col2': 4, 'cnt': 1},
+                          {'col1': 'c', 'col2': 7, 'cnt': 1},
+                          {'col1': 'c', 'col2': 3, 'cnt': 2},
+                          {'col1': 'c', 'col2': 4, 'cnt': 2},
+                          {'col1': 'c', 'col2': 7, 'cnt': 2},
+                          {'col1': 'd', 'col2': 2, 'cnt': 2}])
+
+        self.assertEqual(conn['t1'].windowed(3, 1).windowed(2, 1)
+                         .index('cnt').list(),
+
+                         [{'col1': 'a', 'col2': 4, 'cnt': 0},
+                          {'col1': 'a', 'col2': 5, 'cnt': 0},
+                          {'col1': 'b', 'col2': 1, 'cnt': 0},
+                          {'col1': 'a', 'col2': 5, 'cnt': 0},
+                          {'col1': 'b', 'col2': 1, 'cnt': 0},
+                          {'col1': 'c', 'col2': 3, 'cnt': 0},
+                          {'col1': 'a', 'col2': 5, 'cnt': 1},
+                          {'col1': 'b', 'col2': 1, 'cnt': 1},
+                          {'col1': 'c', 'col2': 3, 'cnt': 1},
+                          {'col1': 'b', 'col2': 1, 'cnt': 1},
+                          {'col1': 'c', 'col2': 3, 'cnt': 1},
+                          {'col1': 'c', 'col2': 4, 'cnt': 1},
+                          {'col1': 'b', 'col2': 1, 'cnt': 2},
+                          {'col1': 'c', 'col2': 3, 'cnt': 2},
+                          {'col1': 'c', 'col2': 4, 'cnt': 2},
+                          {'col1': 'c', 'col2': 3, 'cnt': 2},
+                          {'col1': 'c', 'col2': 4, 'cnt': 2},
+                          {'col1': 'c', 'col2': 7, 'cnt': 2},
+                          {'col1': 'c', 'col2': 3, 'cnt': 3},
+                          {'col1': 'c', 'col2': 4, 'cnt': 3},
+                          {'col1': 'c', 'col2': 7, 'cnt': 3},
+                          {'col1': 'c', 'col2': 4, 'cnt': 3},
+                          {'col1': 'c', 'col2': 7, 'cnt': 3},
+                          {'col1': 'd', 'col2': 2, 'cnt': 3}])
+
     def test_windowed(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -503,6 +647,33 @@ class TestCore(unittest.TestCase):
 
         self.assertEqual(len(conn['t1_1'].list()), 5)
 
+    def test_filter_group(self):
+        conn = Conn(self.dbfile)
+
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].group('col1')\
+            .filter(lambda rs: rs.size() >= 2)
+
+        self.assertEqual(len(conn['t1_1'].list()), 5)
+
+    def test_filter_windowed(self):
+        conn = Conn(self.dbfile)
+
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].windowed(5, 1)\
+            .filter(lambda rs: rs.filter(lambda r: r['col2'] > 3).size() >= 3)
+        self.assertEqual(conn['t1_1'].list(),
+                         [{'col1': 'a', 'col2': 4},
+                          {'col1': 'a', 'col2': 5},
+                          {'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'a', 'col2': 5},
+                          {'col1': 'b', 'col2': 1},
+                          {'col1': 'c', 'col2': 3},
+                          {'col1': 'c', 'col2': 4},
+                          {'col1': 'c', 'col2': 7}])
+
     def test_update(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -633,8 +804,8 @@ class TestCore(unittest.TestCase):
         rs1 = Rows(self.t1)
         rs2 = Rows(self.t2)
         rs1 = rs1.chain([{'col1': 'f', 'col2': 9},
-                         {'col1': 'f', 'col2': 8},
-                         {'col1': 'd', 'col2': 11},
+                        {'col1': 'f', 'col2': 8},
+                        {'col1': 'd', 'col2': 11},
 
                          ])
         rs1 = rs1.update(col4=lambda r: r['col2'] * 20)
@@ -732,8 +903,8 @@ class TestCore(unittest.TestCase):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
 
-        [g1, g2, g3, g4] = [list(xs) for _, xs in conn['t1']\
-            .order('col1, col2 desc').group('col1')._iter()]
+        [g1, g2, g3, g4] = [list(xs) for _, xs in conn['t1']
+                            .order('col1, col2 desc').group('col1')._iter()]
         self.assertEqual(
             g1, [{'col1': 'a', 'col2': 5}, {'col1': 'a', 'col2': 4}]
         )
@@ -792,8 +963,8 @@ class TestCore(unittest.TestCase):
         conn['t1_1'] = conn['t1']\
             .filter(lambda r: r['col2'] > 2)\
             .update(
-                col2=lambda r: r['col2'] + 1,
-                col3=lambda r: r['col1'] + str(r['col2'])
+            col2=lambda r: r['col2'] + 1,
+            col3=lambda r: r['col1'] + str(r['col2'])
         )
         self.assertEqual(conn['t1_1'].list(),
                          [{'col1': 'a', 'col2': 5, 'col3': 'a5'},
